@@ -12,7 +12,6 @@ from time import time
 
 # External:
 from pymining import itemmining
-from blist import sortedset
 
 ###########################################################################
 #                         Association Rule Mining                         #
@@ -57,7 +56,7 @@ def append_rule(
         return
 
     # Finally add the rule, after all those tests.
-    rules.append((left, right, support, confidence, kulc, ir, (1.0 - ir) * kulc))
+    rules.append((left, right, support, (1.0 - ir) * kulc))
 
     # Lookup set, so we don't calculate the same rule more than once.
     known_rules.add((left, right))
@@ -87,7 +86,7 @@ def association_rules(data, min_confidence=0.5, min_support=2, min_kulc=0.66, ma
     :param max_ir: Maximum Imbalance Ratio (from 0 to 1, lower is better)
     :type max_ir: float
     :returns: An iterable with rules.
-    :rtype: [(left, right, support, confidence, kulc, ir, rating), ...]
+    :rtype: [(left, right, support, rating), ...]
     """
     visited, rules, known_rules = set(), deque(), set()
 
@@ -107,9 +106,9 @@ def association_rules(data, min_confidence=0.5, min_support=2, min_kulc=0.66, ma
                 # We copy right, so it does not get harmed during iteration:
                 for item in chain(right, [None]):
                     append_rule(
-                            data, visited, rules, known_rules,
-                            support, left, right, min_confidence,
-                            min_kulc, max_ir
+                        data, visited, rules, known_rules,
+                        support, left, right, min_confidence,
+                        min_kulc, max_ir
                     )
                     if item is not None:
                         cap = [item]
@@ -151,6 +150,13 @@ class History:
         'Iterate over all songs in the history'
         return chain.from_iterable(self.groups())
 
+    def __reversed__(self):
+        'Iterate over all songs in the history in the reverse order, newest first'
+        for group in reversed(self.groups()):
+            for song in reversed(list(group)):
+                yield song
+        # return chain.from_iterable(reversed(self.groups()))
+
     def last_time(self):
         """Gives the last inserted timestamp, or if empty, the current time"""
         # Check if we have a current group:
@@ -234,9 +240,14 @@ class History:
 class ListenHistory(History):
     """A History that holds all recently listened Songs.
     """
-    def __init__(self):
+    def __init__(self, maxlen=10000, max_group_size=5, time_threshold_sec=1200):
         'Sane defaults are chosen for ``History.__init__``'
-        History.__init__(self, maxlen=10000, max_group_size=5)
+        History.__init__(
+            self,
+            maxlen=maxlen,
+            time_threshold_sec=time_threshold_sec,
+            max_group_size=max_group_size
+        )
 
     def frequent_itemsets(self, min_support=2):
         """Mine frequent item sets (FIM) using the RELIM algorithm.
@@ -267,11 +278,63 @@ class ListenHistory(History):
 
 
 class RecommendationHistory(History):
-    """A History that holds all recently given recommendations.
+    """Save made recommendations in order to filter *bad* recommendations.
+
+    *bad* shall be defined as follows: If a recommendation with e.g. the same
+    artist is repeated, even though, the previous recommendation was from this
+    artist it's called *bad*. The same applies for the album or even for the genre.
+
+    In this case we might consider to give a *penalty* to those, so that the same
+    artist only might be recommended within the range of, say, 3 recommendations,
+    and the same album might only be recommended in the range of 5 recommendations.
+
+    The :class:`RecommendationHistory` can do this for any attribute you might
+    think of.
     """
-    def __init__(self):
-        'Sane defaults are chosen for ``History.__init__``'
-        History.__init__(self, maxlen=100)
+    def __init__(self, penalty_map={'artist': 3, 'album': 5}, **kwargs):
+        """
+        .. note::
+
+            This takes the same keyword arguments as :class:`History`, but
+            ``maxlen`` will be silenty ignored.
+
+        :param penalty_map: A mapping from :term:`attribute` s to a penalty count
+                            i.e. the number of recommendations to wait before
+                            allowing it again.
+        """
+        # Hijack the users input, hope he doesn't see this line and gets angry.
+        max_group_size = kwargs.setdefault('max_group_size', 5)
+        kwargs['maxlen'] = max(penalty_map.values()) // max_group_size + 1
+        History.__init__(self, **kwargs)
+
+        self._penalty_map = penalty_map
+
+    def allowed(self, recommendation):
+        """Check if a recommendation is allowed, i.e. if it someting similar
+        was not recently recommended.
+
+        :param recommendation: The recommendation to check for forbidden attributes.
+        :type recommendation: :class:`munin.song.Song`
+        :returns: True if it is allowed, i.e. all possibly forbidden attributes
+                  are long enough ago.
+        """
+        for attribute, penalty in self._penalty_map.items():
+            forbidden = recommendation.get(attribute)
+            if forbidden is None:
+                continue
+
+            for idx, song in enumerate(reversed(self)):
+                if idx >= penalty:
+                    break
+
+                comparable = song.get(attribute)
+                if comparable is None:
+                    continue
+
+                if comparable == forbidden:
+                    return False
+
+        return True
 
 
 ###########################################################################
@@ -307,15 +370,57 @@ class RuleIndex:
         self._max_rules = maxlen or 2 ** 100
         self._rule_list = OrderedDict()
         self._rule_dict = defaultdict(set)
-        self._rule_pool = sortedset(key=_sort_by_rating)
         self._rule_cuid = 0
 
-    def __contains__(self, rule_tuple):
-        'Check if a rule tuple is in the index.'
-        return rule_tuple in self._rule_pool
-
     def __iter__(self):
-        return reversed(self._rule_pool)
+        """Iterate over all rules in a sorted way.
+
+        Requires O(n log n). This could be optimized.
+        """
+        return iter(sorted(self._rule_list.values(), key=lambda rule: 1.0 - rule[-1]))
+
+    def __contains__(self, rule_tuple):
+        'Check if a rule tuple is in the index. Only considers songs in it.'
+        return bool(self.locate(rule_tuple))
+
+    def _locate(self, rule_tuple):
+        """Locate the the first rule_tuple and rule_uid with same left/right.
+
+        :returns: a tuple of (rule_uid, rule_tuple)
+        """
+        first_song = next(iter(rule_tuple[0]))
+        lefts, rights, *_ = rule_tuple
+        candidate_ids = set()
+
+        # Locate all rules that contain a song in the input rule
+        for uid in self._rule_dict.get(first_song, ()):
+            if uid in self._rule_list:
+                candidate_ids.add(uid)
+
+        # Check those for left/right idendity (instead of checking all rules)
+        for uid in candidate_ids:
+            stored_tuple = self._rule_list[uid]
+            stored_lefts, stored_rights, *_ = stored_tuple
+
+            if lefts == stored_lefts and rights == stored_rights:
+                return uid, stored_tuple
+
+            if rights == stored_lefts and lefts == stored_rights:
+                return uid, stored_tuple
+
+        return None, None
+
+    def best(self):
+        """Return the currently best rule (the one with the highest rating)
+
+        Requires linear complexity. This could be optimized.
+
+        :returns: A ruletuple or None if no rule yet in the index.
+        """
+        try:
+            return max(self._rule_list.values(), key=lambda rule: rule[-1])
+        except ValueError:
+            return None
 
     def insert_rules(self, rule_tuples):
         """Convienience function for adding many rules at once.
@@ -336,7 +441,11 @@ class RuleIndex:
                              immediately if index is too large.
         """
         # Step 0: Check if we already know that item.
-        if rule_tuple in self:
+        stored_uid, stored_rule = self._locate(rule_tuple)
+
+        # Update if the rating is better:
+        if stored_rule is not None and rule_tuple[-1] > stored_rule[-1]:
+            self._rule_list[stored_uid] = rule_tuple
             return
 
         # Step 1: Add the affected songs to the index:
@@ -346,13 +455,11 @@ class RuleIndex:
 
         # Step 2: Remember this rule, so we can look it up later.
         self._rule_list[self._rule_cuid] = rule_tuple
-        self._rule_pool.add(rule_tuple)
         self._rule_cuid += 1
 
         # Step 3: Prune the index, if too big.
         if len(self._rule_list) > self._max_rules:
             fst_uid, fst_rule = self._rule_list.popitem(last=False)
-            self._rule_pool.remove(fst_rule)
             if drop_invalid:
                 for uid_set in self._rule_dict.values():
                     uid_set.discard(fst_uid)
@@ -377,7 +484,7 @@ class RuleIndex:
             >>> with rule_index.begin_add_many():
             ...    rule_index.insert_rule(...)
 
-        Calls :func:`drop_invalid` after some time.
+        Calls :func:`drop_invalid` after being done.
         """
         try:
             yield
@@ -400,10 +507,6 @@ class RuleIndex:
                 if not uid in self._rule_list:
                     uid_set.remove(uid)
 
-        # Make sure the pool reflects the current state:
-        self._rule_pool.clear()
-        self._rule_pool.update(self._rule_list.values())
-
 ###########################################################################
 #                               Unit Tests                                #
 ###########################################################################
@@ -443,7 +546,7 @@ if __name__ == '__main__':
                     self.assertTrue('one' in left or 'one' in right)
 
                 # Check the number of items in the index:
-                self.assertEqual(len(self._idx._rule_pool), 10)
+                self.assertEqual(len(self._idx._rule_list), 10)
                 self.assertEqual(len(self._idx._rule_dict), 4)
                 for value in self._idx._rule_dict.values():
                     if setting is False:
@@ -489,6 +592,27 @@ if __name__ == '__main__':
             for group in history.groups():
                 self.assertEqual(len(list(group)), 5)
 
+        def test_recommendation_history(self):
+            history = RecommendationHistory()
+            session = Session('test',
+                {'artist': (None, None, 1), 'album': (None, None, 1)}
+            )
+            fst_song = Song(session, {
+                'artist': 'A',
+                'album': 'B'
+            })
+
+            self.assertEqual(history.allowed(fst_song), True)
+            history.feed(fst_song)
+            self.assertEqual(history.allowed(fst_song), False)
+
+            for expectation in [False, False, False, False, True]:
+                history.feed(Song(session, {
+                    'artist': 'X',
+                    'album': 'Y'
+                }))
+                self.assertEqual(history.allowed(fst_song), expectation)
+
         def test_relim(self):
             history = ListenHistory()
 
@@ -523,11 +647,11 @@ if __name__ == '__main__':
             print()
 
             rules = history.find_rules(itemsets)
-            for left, right, support, confidence, kulc, irat, rating in rules:
-                print('{:>15s} <-> {:<15s} [supp={:> 5d}, conf={:.3f}, kulc={:.5f} irat={:.5f} rating={:.5f}]'.format(
+            for left, right, support, rating in rules:
+                print('{:>15s} <-> {:<15s} [supp={:> 5d}, rating={:.5f}]'.format(
                     str([song.uid for song in left]),
                     str([song.uid for song in right]),
-                    support, confidence, kulc, irat, rating
+                    support, rating
                 ))
 
     unittest.main()
